import cartopy.crs as ccrs

projections = {
    "EuroPP": ccrs.EuroPP,
    "PlateCarree": ccrs.PlateCarree,
    "Mercator": ccrs.Mercator,
    "Robinson": ccrs.Robinson,
    "AlbersEqualArea": ccrs.AlbersEqualArea,
    "LambertConformal": ccrs.LambertConformal,
    "LambertCylindrical": ccrs.LambertCylindrical,
    "Miller": ccrs.Miller,
    "Mollweide": ccrs.Mollweide,
    "Orthographic": ccrs.Orthographic,
    "Sinusoidal": ccrs.Sinusoidal,
    "Stereographic": ccrs.Stereographic,
    "TransverseMercator": ccrs.TransverseMercator,
    "UTM": ccrs.UTM,
    "InterruptedGoodeHomolosine": ccrs.InterruptedGoodeHomolosine,
    "RotatedPole": ccrs.RotatedPole,
    "OSGB": ccrs.OSGB,
    "Geostationary": ccrs.Geostationary,
    "NearsidePerspective": ccrs.NearsidePerspective,
    "LambertAzimuthalEqualArea": ccrs.LambertAzimuthalEqualArea,
    "NorthPolarStereo": ccrs.NorthPolarStereo,
    "OSNI": ccrs.OSNI,
    "SouthPolarStereo": ccrs.SouthPolarStereo
}

