/* Resize logo */
.wy-side-nav-search .logo {
    max-height: 140px;
}

/* Remove extra bottom space below the logo */
.wy-side-nav-search .wy-side-nav-search-logo {
    margin-bottom: 0 !important;
    padding-bottom: 0 !important;
}